#ifndef PRR_H
#define PRR_H

#include "../util/graphs.hpp"
#include "../util/defs.hpp"

namespace cr{


  struct path_info_t{
    // start and end of the path
    edge_p start, end;
    // list of generators: edge points TO the generator
    list<edge_p> generators;
    // list of vertices with pendant Y-graphs on this path
    list<vertex_p> pendantYs;

    // do we have a separator
    bool has_separator;
    vertex_p separator;

    // length is the number of inner edges, that is the number of inner vertices + 1
    uint length;
    // used to invalidate the path, for example when destroying it
    bool valid;

    path_info_t():has_separator(false),length(1),valid(false){}
    bool is_loop(){return start->partner == end->get_reversed()->partner;}
  };


  edge_p skip_ahead_non_intrusive(const uint dfs_id, edge_p& direction, uint *length = NULL);


  // apply Tree Reduction Rules 1-5 from the given leaves upwards
  solution_t apply_prrs(instance& I);

}


ostream& operator<<(ostream& os, const cr::path_info_t& t);

#endif
