#ifndef TRR_H
#define TRR_H

#include "../util/graphs.hpp"
#include "../util/defs.hpp"

namespace cr{
  // update the trr_infos assuming the trr_infos are up to date in the subtree below v
  void update_trr_infos(instance& I, vertex_p v);

  // apply Tree Reduction Rule X to v assuming all TRRs have been applied to the subtree below v
  bool trr14_subtree(instance& I, const vertex_p& v);
  solution_t trr2_subtree(instance& I, const vertex_p& v);
  solution_t trr3_subtree(instance& I, const vertex_p& v);
  solution_t trr5_subtree(instance& I, const vertex_p& v);

  // perofrm the trrs on vertex v and return whether something changed. Store the deleted edges in local_operations
  solution_t perform_trrs(instance& I, const vertex_p& v);

  // update parents trr_infos from child (e := ( parent -> child) )
  void update_trr_infos_from_child(const edge_p& e);

  // Tree Reduction Rule 6 removes caterpillars
  bool trr6(instance& I);

  // applies the tree reduction rules from v upwards after v got thrown out of the cyclic core
  // stops at do_not_cross or a cyclic-core vertex, whichever comes first
  // returns the last found vertex on the cyclic core if there is such a vertex (or any vertex otherwise)
  // cautious: make sure v's subtree_NH are updated before running this!
  solution_t apply_trrs_upwards_after_cut(instance& I, vertex_p& v, const vertex_p do_not_cross);
  solution_t apply_trrs_upwards_after_cut(instance& I, vertex_p& v);

  // apply all trrs to all vertices, return number of deletions done
  // TODO: this is much like graph::update_subtree_NH, try to collapse them into one "bottom-up applicator"
  solution_t apply_trrs(instance& I);
  solution_t apply_trrs(graph& g);
}

#endif
