#ifndef B_VECTOR_HPP
#define B_VECTOR_HPP

#include "branching.hpp"
#include <math.h>

#define PRECISION 5

// branching number code by Joseph, Chuang-Chieh Lin (lincc@cs.ccu.edu.tw or josephcclin@gmail.com)
double CH_POLY(int *A, double var, int n){
	double s = 0;
 	for(int i=0; i != n; i++) s += pow(var,*(A+i));
  return 1-s; 
}

double BranchNumber(int *BV, int n){
  double temp = 0;
  double poly_result = 1;
  for(int d=1; d<=PRECISION; d++) { 
    poly_result = 1;
    while (poly_result > 0) {
      temp += pow(0.1, d);
      poly_result = CH_POLY(BV, temp, n);
    } 
    temp -= pow(0.1, d);
  }
  return (1/temp);
}

float branch_number(const cr::branch_op& bop){
  const std::list<cr::edgelist>& branches(bop.branches);

  if(branches.empty()) return FLT_MAX;
  if(branches.size() == 1) return 1/(float)branches.size();

  int *b_vector = new int[bop.branches.size()];
  uint index = 0;
  for(auto i = branches.begin(); i != branches.end(); ++i)
    b_vector[index++] = i->size();
  float result = (float)BranchNumber(b_vector, branches.size());
  delete[] b_vector;
  return result;
}


#endif
