#include "../util/defs.hpp"
#include "../util/graphs.hpp"
#include "../reduction/trr.hpp"
#include "../reduction/prr.hpp"
#include "../reduction/global.hpp"
#include "bounds.hpp"
#include "branching.hpp"
#include "b_vector.hpp"

#include <algorithm> // for sort
#include <unordered_map>
#include <unordered_set>

//  struct branch_op {
//    branch_type type;
//    // the actual branches
//    list<edgelist> branches;
//    // an _estimation_ on the braching number
//    uint branch_num;
//  };

namespace cr{

  void get_branching_ops(const vertex_p v, branchlist& br){
#warning write me!
  }
  // return true if it contains permanent edges
  bool contains_permanent(edgelist& el){
    for(edgelist::const_iterator e = el.begin(); e != el.end(); ++e)
      if((*e)->is_permanent){
        return true;
      }
    return false;
  }


  // get branching ops for BRR1
  // TODO: this is not optimal! maintain a list of triangles in G instead!
  bool BRR1(const vertex_p v, branchlist& br){
    edgelist non_bridges(v->get_cyclic_neighbors());
    DEBUG2(cout << "this is BRR1 for "<<*v<<" with "<<v->adj_list.size()<< " neighbors"<< endl);
    for(auto i = non_bridges.begin(); i != non_bridges.end(); ++i){
      const edge_p& a(*i);
      for(auto j = i; j != non_bridges.end(); ++j){
        const edge_p& b(*j);
        if(a != b){
          edge_p c = find_edge(a->partner, b->partner);
          if(c != a->partner->adj_list.end()){
            branch_op bo(Triangle);
            if(!a->is_permanent) bo.branches.push_back({a});
            if(!b->is_permanent) bo.branches.push_back({b});
            if(!c->is_permanent) bo.branches.push_back({c});
            DEBUG2(cout << "found 3CYC on "<<*a<<", "<<*b<<", "<<*c<<", non-permanent: "<<bo.branches<<endl);
            if(!bo.branches.empty()){
              br.push_back(bo);
              // one BRR1 for each vertex is enuff, it's not getting better branching numbers anyway
              return true;
            }
          } // end if
        } // end if
      } // end for
    } // end for
    return false;
  } // end function

  // prevent branches containing non-relevant A-bridges
  inline void add_branch(branch_op& bop, const edgelist& el){
    if(!contains_non_relevant_Abridge(el))
      bop.branches.push_back(el);
  }

  edgelist::const_iterator get_nth(const edgelist& el, uint n){
    edgelist::const_iterator i = el.begin();
    while(n){
      ++i;
      --n;
    }
    return i;
  }
  edgelist::const_iterator operator+(const edgelist::const_iterator i, uint n){
    edgelist::const_iterator j = i;
    while(n){
      ++j;
      --n;
    }
    return j;
  }
  edgelist::iterator operator+(const edgelist::iterator i, uint n){
    edgelist::iterator j = i;
    while(n){
      ++j;
      --n;
    }
    return j;
  }

  // return whether BRR2 applies to the v_i vertices
  bool BRR2(const edgelist& vis, const vector<edgelist>& E, branchlist& br){
    unsigned char nonempty_branches = vis.size();
    if(E[0].size() == 1){
      branch_op bop(Claw0);
      add_branch(bop, E[0]);
      if(nonempty_branches > 1) add_branch(bop, E[1]);
      if(nonempty_branches > 2) add_branch(bop, E[2]);

      // edges {e1,e2}
      if(nonempty_branches > 1)
        add_branch(bop, { *(vis.begin()), *(vis.begin()+1) });

      DEBUG2(cout << "BRR2 computed branch op "<<bop<<endl);
      br.push_back(bop);
      return true;
    } else return false;
  }
  // return whether BRR2 applies to the v_i vertices
  bool BRR3(const edgelist& vis, const vector<edgelist>& E, branchlist& br){
    unsigned char nonempty_branches = vis.size();
    if((E[0].size() > 1) && ((nonempty_branches < 2) || (E[1].size() == 1))){
      branch_op bop(Claw1);
      add_branch(bop, E[0]);
      if(nonempty_branches > 1) add_branch(bop, E[1]);
      if(nonempty_branches > 2) add_branch(bop, E[2]);
      // edge {e1}
      add_branch(bop, {vis.front()});

      // edge {e2,e3}
      if(nonempty_branches > 2)
        add_branch(bop, {*(vis.begin()+1), *(vis.begin()+2) });
    
      DEBUG2(cout << "BRR3 computed branch op "<<bop<<endl);
      br.push_back(bop);
      return true;
    } else return false;
  }
  // return whether BRR2 applies to the v_i vertices
  bool BRR4(const edgelist& vis, const vector<edgelist>& E, branchlist& br){
    unsigned char nonempty_branches = vis.size();
    // only applies if there are at least two relevant branches
    if(nonempty_branches < 2) return false;
    if((E[1].size() > 1) && ((nonempty_branches < 3) || (E[2].size() == 1))){
      branch_op bop(Claw2);
      add_branch(bop, E[0]);
      add_branch(bop, E[1]);
      if(nonempty_branches > 2) add_branch(bop, E[2]);
      // edge {e1}
      add_branch(bop, {vis.front()});

      // edge {e2}
      add_branch(bop, {*(vis.begin()+1)});

      DEBUG2(cout << "BRR4 computed branch op "<<bop<<endl);
      br.push_back(bop);
      return true;
    } else return false;
  }
  // return whether BRR2 applies to the v_i vertices
  bool BRR5(const edgelist& vis, const vector<edgelist>& E, branchlist& br){
    unsigned char nonempty_branches = vis.size();
    // only applies if there are at least two relevant branches
    if(nonempty_branches < 3) return false;
    if(E[2].size() > 1){
      branch_op bop(Claw3);
      add_branch(bop, E[0]);
      add_branch(bop, E[1]);
      add_branch(bop, E[2]);
      // edge {e1}
      add_branch(bop, {vis.front()});

      // edge {e2}
      add_branch(bop, {*(vis.begin()+1)});

      // edge {e3}
      add_branch(bop, {*(vis.begin()+2)});

      DEBUG2(cout << "BRR5 computed branch op "<<bop<<endl);
      br.push_back(bop);
      return true;
    } else return false;
  }








  // note:
  // in A bridges, one endpoint is the parent of the other!
  // a B-bridge is a bridge that is not an A-bridge

  // return whether e points to an eligible branching partner
  bool is_eligible_branching_partner(const edge_p& e){
    DEBUG2(cout<<"testing "<<*e<<endl);
    // branching into y-graphs on deg2-paths is not good! (e is non-relevant Abridge)
    const vertex_p& v(e->get_reversed()->partner);
    if(e->partner->degree() == 1) return false;
    DEBUG2(cout<<"passed 1st test: "<<*(e->partner)<<" is not a leaf"<<endl);
    if((v->non_bridge_degree() == 2) && (!v->trr_infos.ygraphs.empty()))
      if(v->trr_infos.ygraphs.front() == e) return false;
    DEBUG2(cout<<"passed 2st test: "<<*e<<" is not a relevant A-bridge"<<endl);
    if(e->partner->is_incident_to_Bbridge()) return false;
    DEBUG2(cout<<"passed 3nd test: "<<*(e->partner)<<" is not incident to B-bridges"<<endl);
    return true;      
  }

  void get_non_bridge_branching_partners(const vertex_p& v, edgelist& el){
    DEBUG2(cout << "getting non-bridge branching partners of "<<*v<<endl);
    for(edge_p e = v->adj_list.begin(); e != v->adj_list.end(); ++e)
      if(!e->is_bridge){
        if(is_eligible_branching_partner(e))
          el.push_back(e);
      }

    DEBUG2(cout << "found " << (el.empty()?"none":"these: ") << el << endl);
  }
  // select three branching partners to fulfill the following conditions:
/*
 *  (1) deg(v1) = 2 ==> deg(v2) = 2 and deg(v2) = 2 ==> deg(v3) = 2.
 *  (2) if deg(v1) = 2, then deleting {v, v1} does not make {v, v2} a bridge, and
 *  (3) if deg(v1) > deg(v2) = 2, then deleting {v, v2} does not make {v, v3} a bridge.
 */
  bool greater_than(const edge_p& a, const edge_p& b){
    return a->partner->degree() > b->partner->degree();
  }
  // if deleting *i makes *j a bridge, then replace *j with *(++j)
  void bring_in_order(edgelist& el, const edgelist::iterator i, edgelist::iterator j){
    const vertex_p& v((*i)->get_reversed()->partner);
    edge_p e = *i;
    // find out whether going along (v,v2) returns to v with (v3,v)
    DEBUG2(cout << "checking whether killing "<<**i<<" makes "<<**j<<" a bridge"<<endl);
    skip_ahead_non_intrusive(v->dfs_id, e);
    if(e->get_reversed() == *j){
      DEBUG2(cout << "indeed so! Maybe there is more on the edgelist: "<<el<<" ("<<el.size()<<" items)"<<endl);
      DEBUG2(cout << "i is "<<**i<<endl);
      DEBUG2(cout << "j is "<<**j<<endl);
      // swap the second and third items on el
      // NOTE: if there are no more items on the list, then v4 is incident to a B-bridge,
      // and therefore, not an eligible branching partner - in this case, we cannot branch here
      edgelist::iterator k = j;
      ++k;
      if(k == el.end()){
        el.clear();
      } else {
        swap(*j, *k);
      }
    }
  }
  // TODO: preferably get ptwos as branching partners
  bool select_branching_partners(edgelist& el){
    const vertex_p& v(el.front()->get_reversed()->partner);
    uint ptwos = v->trr_infos.ptwos.size();

    // if there is not enuff to branch on, return false
    DEBUG2(cout << "totalling "<<el.size()+ptwos<<" possible branching partners"<<endl);
    if(el.size() + ptwos < 3) return false;

    DEBUG2(cout << "sorting by degree"<<endl);
//    sort(el.begin(), el.end(), greater_than);
    el.sort(greater_than);
    DEBUG2(cout << "new order: "<<el<<endl);

    if(v->trr_infos.ptwos.empty()){
      
      edgelist::iterator i = el.begin();
      // ensure (2)
      if(el.front()->partner->degree() == 2){
        DEBUG2(cout<<"ensuring (2) for "<<**i<<endl);
        edgelist::iterator j = i;
        bring_in_order(el, i, ++j);
      };
      // bring_in_order can determine that the branching partners are invalid. In this case, return
      if(el.empty()){
        DEBUG2(cout << "branching partners were not valid, returning with empty edgelist"<<endl);
        return false;
      }
      ++i;
      // ensure (3)
      if((el.front()->partner->degree() > 2) &&
        ((*i)->partner->degree() == 2)){
        DEBUG2(cout<<"ensuring (3) for "<<**i<<endl);
        edgelist::iterator j = i;
        bring_in_order(el, i, ++j);
      }
      // bring_in_order can determine that the branching partners are invalid. In this case, return
      if(el.empty()){
        DEBUG2(cout << "branching partners were not valid, returning with empty edgelist"<<endl);
        return false;
      }

      // erase all remaining items
      el.erase(el.begin()+3, el.end());
    } else { // v has ptwos, take those first, then fill up with any partner
      DEBUG2(cout << *v << " has "<<ptwos<<" P2s, using those"<<endl);
      el.erase(el.begin(), el.begin() + (el.size() - (3 - ptwos)));
    }
    DEBUG2(cout<<"everything OK, branching partners are now "<<el<<endl);
    return true;
  }
  // returns true if there is a path from u to v avoiding do_not_cross and returns the last_edge on the path
  bool find_path_avoiding(const vertex_p& u, const vertex_p& v, const edge_p& do_not_cross, const size_t dfs_id, edge_p& last_edge){
    if(u->dfs_id == dfs_id) return false;

    u->dfs_id = dfs_id;
    for(edge_p e = u->adj_list.begin(); e != u->adj_list.end(); ++e) if(e != do_not_cross){
      if(e->partner == v){
        last_edge = e;
        DEBUG2(cout << "heureka, found path from "<<*u<<" to "<<*v<<" avoiding "<<*do_not_cross<<". Last edge is "<<*e<<endl);
        return true;
      } else if(find_path_avoiding(e->partner, v, do_not_cross, dfs_id, last_edge)) return true;
    }
    return false;
  }

  
  // new version of compute_Ei that returns a set containing _all_ edges incident to vi except {v,vi}
  void compute_Ei(graph& g, const edge_p& ei, edgelist& el){
    const vertex_p& vi(ei->partner);
    DEBUG2(cout << "finding E_i for "<<*ei<<" (non-bridge-degree of "<<*vi<<": "<<vi->non_bridge_degree()<<")"<<endl);

    for(edge_p e = vi->adj_list.begin(); e != vi->adj_list.end(); ++e)
      if(e != ei->get_reversed())
        el.push_back(e);
  }

  
  void compute_Ei_old(graph& g, const edge_p& ei, edgelist& el){
    const vertex_p& vi(ei->partner);
    const vertex_p& v(ei->get_reversed()->partner);
    DEBUG2(cout << "finding E_i for "<<*ei<<" (non-bridge-degree of "<<*vi<<": "<<vi->non_bridge_degree()<<")"<<endl);

    if(vi->non_bridge_degree() == 0){
      for(edge_p q = vi->adj_list.begin(); q != vi->adj_list.end(); ++q)
        if(q->get_reversed() != ei){
          el.push_back(q);
          if(el.size() == 2) return;
        }
      return;
    }

    edge_p e = vi->adj_list.end();
    if(vi->non_bridge_degree() > 2){
      // find a path from v to vi avoiding {v,vi} and the last edge on that path
      find_path_avoiding(v, vi, ei, g.get_dfs_id(), e);
      e = e->get_reversed();
    } else {
      e = vi->first_non_bridge_neighbor_except(v);
    }
    
    el.push_back(e);

    // select at most one other non-bridge neighbor of vi for Ei
    for(edge_p q = vi->adj_list.begin(); q != vi->adj_list.end(); ++q)
      if((q != e) && (q != ei->get_reversed())){
        if((vi->non_bridge_degree() == 2) || !q->is_bridge){
          el.push_back(q);
          break;
        }
    }
  }



  // get branching ops for BRR2
  bool BRR2_to_5(graph& g, const vertex_p& v, branchlist& br){
    edgelist branch_partners;
    if(v->nldeg() < 3) return false;
    DEBUG2(cout << "This is BRR2-5 for "<<*v<<endl);
    // first, get the branching partners
    get_non_bridge_branching_partners(v, branch_partners);
    // if we can't get any branching partners, then just return
    if(branch_partners.empty()) return false;
    // then, select the ones that are good
    DEBUG2(cout << "selecting correct branching partners from "<<branch_partners<<endl;);
    if(!select_branching_partners(branch_partners)) return false;

    vector<edgelist> Eis;
    for(edgelist::const_iterator e = branch_partners.begin(); e != branch_partners.end(); ++e){
      edgelist el;
      // compute the sets E_i and store em in Eis
      compute_Ei(g, *e, el);
      Eis.push_back(el);
    }
    DEBUG2(cout << "going into branching rules with "<<branch_partners<<" and "<<Eis<<endl;);

    if(BRR2(branch_partners, Eis, br)) return true;
    if(BRR3(branch_partners, Eis, br)) return true;
    if(BRR4(branch_partners, Eis, br)) return true;
    if(BRR5(branch_partners, Eis, br)) return true;

    FAIL("epic fail in BRR2_to_5");
  }



  bool get_best_branch_op(graph& g, branch_op& bo){
    branchlist br;
    // check BR1 first
    for(vertex_p v = g.vertices.begin(); v != g.vertices.end(); ++v) if(v->is_on_cycle()){
      if(BRR1(v, br))
        if(br.back().branches.size() == 1){
          bo = br.back();
          if(contains_permanent(bo.branches.front())) bo.branches.clear();
          return true; // if we found a reduction, we're happy
        }
    }
    DEBUG2(cout << "done applying BRR1"<<endl);
    // if BRR1 is not applicable to G, try BRR 2-5
    for(vertex_p v = g.vertices.begin(); v != g.vertices.end(); ++v) if(v->is_on_cycle()){
      if(BRR2_to_5(g, v, br))
        if(br.back().branches.size() == 1){
          bo = br.back();
          return true; // if we found a reduction, we're happy
        }
    }
    if(br.empty()){
      DEBUG2(cout << "couldn't branch: no branchable vertices"<<endl);
      return false;
    }
    for(auto bop = br.begin(); bop != br.end(); ++bop){
      for(auto el = bop->branches.begin(); el != bop->branches.end(); ++el){
        if(contains_permanent(*el)) el = bop->branches.erase(el);
        if(el == bop->branches.end()) break;
      }
      if(bop->branches.size() <= 1){
        bo = *bop;
        return true; // if we found a reduction, we're happy
      }
    }

    branchlist::iterator best_op;
    float best_bnum = FLT_MAX;
    for(auto bop = br.begin(); bop != br.end(); ++bop){
      DEBUG2(cout << "branching num for "<<*bop<<":"<<endl);
      float bnum = branch_number(*bop);
      DEBUG2(cout<<bnum<<endl);
      if(bnum < best_bnum){
        best_bnum = bnum;
        best_op = bop;
      }
    }
    bo = *best_op;
    DEBUG2(cout << "best branching op is "<<bo<<" with bnum "<<best_bnum<<endl);
    return true;
  }

  // main function solving the problem!
  // TODO: we copy the graph, this is inefficient! improve!
  solution_t run_branching_algo(instance& I, size_t& search_tree_nodes, const solv_options& opts, const uint depth){
    // keep track of the search tree size
    search_tree_nodes++;
    DEBUG5(if(search_tree_nodes % 10000 == 0) cout << "currently at "<< search_tree_nodes<<" nodes"<<endl;);
    
    solution_t sol;

    // [1.] apply preprocessing
    DEBUG4(cout << "=== Phase 1 (depth "<<depth<<"): TRRs ===== (k = "<<I.k<<")"<<endl);
    sol += apply_trrs(I);
    DEBUG2(I.g.write_to_stream(std::cout));
    DEBUG4(cout << "=== Phase 2 (depth "<<depth<<"): PRRs ("<< sol.size() <<" dels, k = "<<I.k<<") ====="<<endl);
    sol += apply_prrs(I);
    DEBUG2(cout << "welcome back to apply_prrs - sol: "<<sol<<endl);

    // if preprocessing solved I, then return success
    if(I.g.vertices.empty() && !(I.k < 0)) return sol;
   
    // return failure if preprocessing already took all the operations
    // we do another check later, but here, we can avoid computing the lower bound
    if(I.k <= 0) return solution_t();
 
    // [2.] get a lower bound
    // prepare the graph (it shouldn't be required for finding connected components, but we'll need it later anyways)
    // the (better) lower bound by star packing is slower, so just use it every once in a while
    int lower_bnd = compute_lower_bound(I.g, opts, depth);

    DEBUG4(cout << "=== Phase 3 (depth "<<depth<<"): compare budget (" << I.k <<") to lower bound ("<< lower_bnd <<") ====="<<endl);

    // if the lower bound already exceeds our budget then give up
    if(lower_bnd > I.k) {I.k = -1; return solution_t();}

    // [3.] split off connected components if possible
    DEBUG4(cout << "=== Phase 4 (depth "<<depth<<"): global reduction rules & detect connected components ===" << endl);
    //apply_split_rule(I); // -- this doesn't happen... at all

    // compute the cc number
    I.g.mark_bridges();

    if(I.g.cc_number > 1){
      instance Iprime;
      DEBUG2(cout << "detected " << I.g.cc_number << " components, splitting g"<<endl;);
      
      // split off a connected component from I
      split_off_component(I.g, Iprime.g);

      // recurse for both components, summing up solutions (smaller one first)
      if(I.g.vertices.size() > Iprime.g.vertices.size()){
        sol += run_branching_algo(I, search_tree_nodes, opts, depth+1);
        // if there was not enough budget to solve I, return failure
        if(!I.g.vertices.empty() || (I.k < 0)) {I.k = -1; return solution_t();}
        // otherwise, use the remaining budget for the split-off component
        Iprime.k = I.k;
        sol += run_branching_algo(Iprime, search_tree_nodes, opts, depth+1);
        // if there was not enough budget to solve this component, then return failure
        if(!Iprime.g.vertices.empty() || (Iprime.k < 0)) {I.k = -1; return solution_t();}
      } else {
        Iprime.k = I.k;
        sol += run_branching_algo(Iprime, search_tree_nodes, opts, depth+1);
        // if there was not enough budget to solve Iprime, return failure
        if(!Iprime.g.vertices.empty() || (Iprime.k < 0)) {I.k = -1; return solution_t();}
        // otherwise, use the remaining budget for the split-off component
        I.k = Iprime.k;
        sol += run_branching_algo(I, search_tree_nodes, opts, depth+1);
        // if there was not enough budget to solve this component, return failure
        if(!I.g.vertices.empty() || (I.k < 0)) {I.k = -1; return solution_t();}
      }
      // if all went well, return success
      return sol;
    }

    // [4.] start branching
    DEBUG4(cout << "=== Phase 5 (depth "<<depth<<"): branchings ("<< sol.size() <<" dels) ====="<<endl);
    DEBUG3(I.g.write_to_stream(std::cout));

    // try to apply the B-bridge rule, which is possible if and only if G has B-bridgs
//    sol += apply_Bbridge_rule(I);
    // if the splitting rule solved the instance (positivly or negatively, return the result)
    if(I.k < 0) return solution_t();
    if(I.g.vertices.empty()) return sol;

    // do the actual branching: first, get a good (the BEST! ^^) branching operation
    branch_op bo;
    if(get_best_branch_op(I.g, bo)){
      // we may be able to mark edges permanent if deleting them fails to get a better solution
      edgelist permanents;

      switch(bo.branches.size()){
        case 0:
          // if there is an empty branch, this means that all
          // branch-edges are permanent and, hence, we have already
          // seen an optimal solution, so no need to continue
          return solution_t();

        case 1:
          // if we have only one branch, then its a reduction rule
          // and we don't need to copy the graph
          DEBUG3(cout << "only one branch: "<<bo.branches<<" - wont copy the graph"<<endl);

          // if the deletions of this branch are already too much, then return failure
          if((int)(bo.branches.front().size()) > I.k) return solution_t();
          
          // otherwise, happily apply the branch
          I.delete_edges(bo.branches.front(), sol);
          // and recurse
          sol += run_branching_algo(I, search_tree_nodes, opts, depth+1);
          return sol;

        default:
          solution_t min_sol;
          int known_solution = I.k + 1;
          // for each branch in the branch list (which doesn't exceed the budget):
          for(auto el = bo.branches.begin(); el != bo.branches.end(); ++el) if((int)(el->size()) <= min(I.k, known_solution - 1)){
            DEBUG2(cout << "depth " << depth << " branch: "<<*el<<endl);
            solution_t solprime;
            // save the first edge of el in case we need to mark it permanent
            edge_p to_be_permanent(el->front());
            // make a copy Iprime of I
            instance Iprime(I, &(*el));
            // we only need to find solutions that are better than what we have
            Iprime.k = min(I.k, known_solution - 1);
            // delete the edges of this branch
            Iprime.delete_edges(*el, solprime);
            // and recurse
            solprime += run_branching_algo(Iprime, search_tree_nodes, opts, depth+1);

            // if we were successfull, save the solution and mark edges permanent for the next branch
            if(Iprime.g.vertices.empty() && !(Iprime.k < 0)){
              DEBUG2(cout << solprime << " is indeed a valid solution and its size is " << solprime.size()<<endl);
              min_sol = solprime;
              // we've found a solution that should be smaller than known_solution
              known_solution = solprime.size();
              // recheck Sud05, but I think we can only mark edgesets of size one permanent!
              if(el->size() == 1) {
                permanents.push_back(to_be_permanent);
                to_be_permanent->mark_permanent();
              }
            }
            DEBUG2(cout << "depth "<<depth<<": current min solution is "<<min_sol << " most recent: "<<solprime << " now searching for solutions of size " << min(I.k, known_solution - 1) <<endl);
          }
          DEBUG2(cout << "depth "<<depth<<": done applying branching, we have "; if(min_sol.empty()) cout << "no solution"; else cout << " a solution: "<<min_sol; cout<< endl);
          // if no solution was found, return failure, otherwise merge the minimum solution into sol and clear the graph
          if(min_sol.empty()) return solution_t(); else{
            I.g.clear();
            sol += min_sol;
          }
          // everything went well, so return the solution
          DEBUG3(cout << "returning with "<<search_tree_nodes<<" search_tree_nodes and solution "<< sol<<endl);
          return sol;
      }
    } else {
      // no branchable vertices have been found, so apply the final RR and recurse
      sol += apply_final_RR(I);
      sol += run_branching_algo(I, search_tree_nodes, opts, depth+1);

      DEBUG3(cout << "returning with "<<search_tree_nodes<<" search_tree_nodes "<<endl);
      return sol;
    }
  }





}; // end namespace

ostream& operator<<(ostream& os, const cr::solv_options& opts){
  return os << " 1. slow lower bound computation each " << opts.slow_lower_bound_layers_wait << " layers"<<endl;
}

ostream& operator<<(ostream& os, const cr::branch_op& bop){
  os << "[";
  switch(bop.type){
    case cr::Triangle:
      os << "Tri: ";
      break;
    default:
      os << "Claw"<< bop.type-1<<":";
  }
  os << bop.branches << "]";
  return os;
}



