#ifndef BRANCHING_HPP
#define BRANCHING_HPP

#include <list>
#include <vector>
#include "../util/graphs.hpp"

namespace cr{
  enum branch_type {Triangle, Claw0, Claw1, Claw2, Claw3};

  struct solv_options{
    uint fast_lower_bound_layers_wait;
    uint slow_lower_bound_layers_wait;
  };
  const solv_options default_opts = {
    1, // fast_lower_bound_layers_wait
    8, // slow_lower_bound_layers_wait
  };

  struct branch_op {
    branch_type type;
    // the actual branches
    list<edgelist> branches;

    // constructors
    branch_op(const branch_type t):type(t),branches(){};
    branch_op() {};
  };
  typedef list<branch_op> branchlist;

  // get branching opportunities for a vertex
  void get_branching_ops(const vertex_p v, branchlist& br);

  // get branching ops for BRR1
  void BRR2_to_5(const vertex_p v, branchlist& br);

  // apply the final reduction rule when nothing else is applicable anymore
  solution_t apply_final_RR(instance& I);

  // run the complete branching recursively and return the number of operation it took
  solution_t run_branching_algo(instance& I, size_t& search_tree_nodes, const solv_options& opts = default_opts, const uint depth = 0);
 
  // inlines
  inline bool contains_Abridge(const edgelist& E){
    for(edgelist::const_iterator i = E.begin(); i != E.end(); ++i)
      if((*i)->is_Abridge()) return true;
    return false;
  }
  inline bool contains_non_relevant_Abridge(const edgelist& E){
    for(edgelist::const_iterator i = E.begin(); i != E.end(); ++i)
      if((*i)->is_Abridge() && !(*i)->is_relevant_Abridge()) return true;
    return false;
  }

}


ostream& operator<<(ostream& os, const cr::solv_options& opts);
ostream& operator<<(ostream& os, const cr::branch_op& bop);

#endif
